# Client settings
BASE_URL = "https://api.mopinion.com"
TOKEN_PATH = "/token"
LATEST_VERSION = "2.0.0"

# Some settings for dataclasses
VERBOSITY_lEVELS = ["quiet", "normal", "full"]
ITERATE_VERBOSITY_lEVELS = ["normal", "full"]
ALLOWED_METHODS = ["get", "post", "put", "delete", "options"]
VERSIONS = ["1.18.14", "2.0.0"]
CONTENT_NEGOTIATIONS = ["application/json", "application/x-yaml"]
